% Enhanced slrn functions.                  % -*- mode: slang; mode: fold -*-
%
% Copyright 2000-2018 J.B. Nicholson
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program; if not, write to the Free Software
%    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%
% Author: J.B. Nicholson <jbn@forestfield.org>
% Summary: A variety of drop-in (and almost drop-in) replacement macros to
%          enhance slrn's functions.
% Version: 2.00
% License: GNU GPL v2+
% Last tested with: slrn 1.0.3 & S-Lang 2.3.1
%
% To install: copy this file to your slrn macros folder (perhaps
% ~/.slrn) and add the following to your slrnrc:
%
%   set macro_directory ".slrn"
%   interpret "enhanced_slrn-2.0.sl"
%
% The first line (set macro_directory) tells slrn where to find macro
% files such as this one. The second line (interpret) reads this file
% into slrn.
%
% History:
% 1.0:  Released.
% 1.01: Removed Enhanced->split()
% 2.0:  Added CursorMovementError exception
%       Added catchup_following()
%       Added uncatchup_following()
%       Added undelete_thread()
%       Added invert_read_thread()
%       Removed reverse_group_search(); slrn's group_search_backward can do this job.

implements ("Enhanced");

try
{
   new_exception ("CursorMovementError", ApplicationError, "Unexpected inability to move cursor!");
}
catch AnyError:
{
   % There is no code here because nothing needs to be done if this
   % exception already exists.
}

% Functions typically used in other functions, programs. They can be
% bound to keys but will need to pop all returned values.

static define version ()                                     { return (2.0); }
static define simple_group_search (desired_group) %{{{
% SYNOPSIS
%   Place the cursor at the intended group or don't move the cursor.
% 
% USAGE
%   Integer_Type Enhanced->simple_group_search (String_Type newsgroup)
% 
% DESCRIPTION
%   This function is a wrapper for slrn's group_search() which leaves the
%   cursor at the newsgroup specified or leaves the cursor where it
%   started.  Unlike slrn's group_search(), simple_group_search() only
%   matches group names. This function is useful when you don't know
%   if the group you want is in the group list.
% 
% EXAMPLE
% 
%   if (Enhanced->simple_group_search ("alt.lang.s-lang"))
%     {
%        message ("Here is the S-Lang group.");
%     }
%   else
%     {
%        message ("I couldn\'t find the S-Lang group.");
%     }
%
% NOTES
%   slrn's add_group function will position the cursor at the group you
%   specify faster than any macro can do so iteratively. But add_group
%   will also find hidden groups (and unhide them).
%
% SEE ALSO
%   add_group
{
   variable starting_group = current_newsgroup ();

   if (starting_group == desired_group) return (1);
   ifnot (group_search (desired_group)) return (0);

   % Found a candidate group, but this match might be a match on a
   % substring or a group description instead of the group name.
   variable identify_loop_point_group = current_newsgroup ();
   if (identify_loop_point_group == desired_group) return (1);
   while (group_search (desired_group))
     {
        switch (current_newsgroup ())
          { case (desired_group): return (1); }
          { case (identify_loop_point_group):
	     % Back where we started.
             set_input_string (starting_group);
             call ("add_group");
             return (0);
          }
     }
}

%}}}
static define header_prev_unread () %{{{
% SYNOPSIS
%   Search backwards for an unread article.
% 
% USAGE
%   Integer_Type Enhanced->header_prev_unread ()
%
% DESCRIPTION
%   This function complements slrn's header_next_unread() by performing the
%   same search in reverse.  Instead of finding an unread article below the
%   cursor, this function finds the nearest unread article above the cursor.
%   
%   0 is returned if no such article is found, non-zero otherwise.  If an
%   unread article is found the cursor is left at that unread article.
%   Otherwise the cursor is returned to the place where the search began.
%   
% SEE ALSO
%   header_next_unread()
{
   if (is_group_mode ()) throw UsageError, _function_name () + " not found in current keymap.";

   variable
     t_size = 0,
     starting_article = extract_article_header ("Message-ID");

   while (header_up (1))
     {
        if (is_thread_collapsed ())
          {
             t_size = thread_size () - 1;
             uncollapse_thread ();

	     % Move to end of thread.
	     () = header_down (t_size);

	     % Skip read articles within thread.
	     loop (t_size)
               {
                  ifnot (get_header_flags () & HEADER_READ) return (1);
                  () = header_up (1);
               }
             collapse_thread ();
          }

        % handle the head of the thread
        ifnot (get_header_flags () & HEADER_READ) return (1);
     }

   () = locate_header_by_msgid (starting_article, 0);
   message ("No previous unread articles.");
   return (0);
}
%}}}

% Functions typically bound to keys for interactive use. These
% functions return no value.

static define prev_high_score () %{{{
% SYNOPSIS
%   Search up the header list for a high-scoring article.
% 
% USAGE
%   Enhanced->prev_high_score ()
%
% DESCRIPTION
%   This function complements slrn's next_high_score functio by
%   performing the same search in reverse.  Instead of moving the
%   cursor down the header list to find a high-scoring article, the
%   cursor is moved up the header list to find a high-scoring article.
%   
%   Returns 0 if no previous high scoring article is found, 1
%   otherwise.
%   
% SEE ALSO
%   next_high_score
{
   if (is_group_mode ()) throw UsageError, _function_name () + " not found in current keymap.";

   variable
     t_size = 0,
     starting_article = extract_article_header ("Message-ID");

   while (header_up (1))
     {
        if (is_thread_collapsed ())
          {
             t_size = thread_size () - 1;
             uncollapse_thread ();

	     % Move to end of thread.
             () = header_down (t_size);  

	     % Skip non-high score articles within thread.
	     loop (t_size)               
               {
                  if (get_header_flags () & HEADER_HIGH_SCORE)
		    {
		       if (_is_article_visible () == 1) call ("scroll_dn");
		       return (1);
		    }
                  () = header_up (1);
               }
             collapse_thread ();
          }

        % Check the head of the thread.
        if (get_header_flags () & HEADER_HIGH_SCORE)
	  {
	     if (_is_article_visible () == 1) call ("scroll_dn");
	     return (1);
	  }
     }

   () = locate_header_by_msgid (starting_article, 0);
   message ("No more high scoring articles.");
   return (0);
}
%}}}
definekey ("() = Enhanced->prev_high_score",  "@",    "article");

static define catchup_following () %{{{
{
   % Mark the current article and all articles after as read. This
   % implements the opposite of the catchup intrinsic function.

   if ((get_variable_value ("confirm_actions") & 1) &&
       ('N' == toupper (get_response ("yYnN", "Mark all articles from here to the end of the list as read? [\001Y]es, [\001N]o"))))
     return;

   variable
     prefix_argument = get_prefix_arg(),
     starting_article = extract_article_header("Message-ID"),
     
     % Used in marking entire threads read with delete_thread intrinsic.
     last_article = "";

   reset_prefix_arg();
   EXIT_BLOCK 
     {
	if (prefix_argument != -1) set_prefix_argument (prefix_argument);
	() = locate_header_by_msgid (starting_article, 0);
     }

   % Mark all subsequent threads as read using an intrinsic.
   do
     {
	last_article = extract_article_header("Message-ID");
	call("delete_thread");
     }
   while (last_article != extract_article_header("Message-ID"));
   % Try to get rid of the message from the last call() above.
   message ("");
}
%}}}
definekey ("Enhanced->catchup_following",     "\eC",  "article");

static define undelete_thread () %{{{
  % SYNOPSIS
  %   Marks a thread (or subthread) unread and leaves the cursor at
  %   the next read article, uncollapsing a thread if necessary. This
  %   implements the opposite of the delete_thread intrinsic function.
  %
  % USAGE
  %   Void undelete_thread()
  % 
  % SEE ALSO
  %   delete_thread intrinsic function
{
   
   variable
     starting_article = extract_article_header ("Message-ID"),
     t_size = thread_size (),
     thread_was_collapsed = is_thread_collapsed (),
     i = 0;
   
   uncollapse_thread ();
   for (i = 1; i <= t_size; i++)
     {
	set_header_flags (get_header_flags () & ~HEADER_READ);
	if (i < t_size) { ifnot (header_down (1)) throw (CursorMovementError); }
     }
   if (thread_was_collapsed) collapse_thread ();
   () = locate_header_by_msgid (starting_article, 0);
   
   % Leave the cursor at the next read article (and expand the thread
   % this article is in if that article is not the head of the thread).
   % 
   % The task is broken up into two major parts: looking for a read
   % article starting somewhere inside a thread, and looking for a
   % read article starting from an article at the head of a thread.
   
   % Find the next read article somewhere inside the thread assuming
   % the cursor is not at the thread-head article.

   % Go to the end of the thread and record that article's Message-ID.
   collapse_thread ();
   uncollapse_thread ();
   () = header_down (thread_size () - 1);
   variable thread_end_article = extract_article_header ("Message-ID");

   % From the starting article until the end of the thread: find a
   % read article, or tell the user there aren't any read articles.
   () = locate_header_by_msgid (starting_article, 0);
   forever
     {
	if (get_header_flags () & HEADER_READ) return;
	if (extract_article_header ("Message-ID") != thread_end_article) break;
	ifnot (header_down (1)) 
	  {
	     message ("No following read articles.");
	     () = locate_header_by_msgid (starting_article, 0);
	     return;
	  }
     }
   if (thread_was_collapsed) collapse_thread ();

   % Move the cursor to the next thread. Assume cursor is at a 
   % thread-heading article.
   ifnot (header_down (1)) 
     {
	message ("No following read articles.");
	() = locate_header_by_msgid (starting_article, 0);
	return;
     }

   % This should handle finding a read article starting at a
   % thread-heading article.
   do
     {
	if (get_header_flags () & HEADER_READ) return;
	thread_was_collapsed = is_thread_collapsed ();
	t_size = thread_size ();
	i = 1;
	uncollapse_thread ();
	for (i = 1; i <= t_size; i++)
	  {
	     if (get_header_flags () & HEADER_READ) return;
	     if (i < t_size) { ifnot (header_down (1)) throw (CursorMovementError); }
	  }
	if (thread_was_collapsed) collapse_thread ();
     }
   while (header_down (1));
   message ("No following read articles.");
   () = locate_header_by_msgid (starting_article, 0);
}
%}}}
definekey ("Enhanced->undelete_thread",       "\eD",  "article");

static define uncatchup_following () %{{{
{
   % Mark all articles unread from the current article to the end of
   % the article list. This implements the opposite of the uncatchup
   % intrinsic function.

   if ((get_variable_value ("confirm_actions") & 1) &&
       ('N' == toupper (get_response ("yYnN", "Mark all articles from here to the end of the list as unread? [\001Y]es, [\001N]o"))))
     return;

   undelete_thread ();

   variable
     thread_was_collapsed = 0,
     t_size = 0,
     i = 0;
   do
     {
	thread_was_collapsed = is_thread_collapsed ();
	t_size = thread_size ();
	i = 1;
	uncollapse_thread ();
	for (i = 1; i <= t_size; i++)
	  {
	     set_header_flags (get_header_flags () & ~HEADER_READ);
	     if (i < t_size) { ifnot (header_down (1)) throw (CursorMovementError); }
	  }
	if (thread_was_collapsed) collapse_thread ();
     }
   while (header_down (1));
}
%}}}
definekey ("Enhanced->uncatchup_following",   "\e^U", "article");

static define invert_read_unread () %{{{
{
   % All read articles will become unread articles, and all unread
   % articles will become read articles.

   % We don't want any prefix arguments to interfere with any call()s
   % we make. So we record the prefix argument, reset the prefix
   % argument, and restore the prefix argument at the end (in the
   % EXIT_BLOCK).
   variable
     prefix_arg = get_prefix_arg(),
     starting_article = extract_article_header ("Message-ID"),
     thread_was_collapsed = 0,
     i = 0,
     t_size = 0;

   reset_prefix_arg();
   EXIT_BLOCK
     {
	% Restore the state we modified.
	if (prefix_arg != -1) set_prefix_argument (prefix_arg);
	() = locate_header_by_msgid (starting_article, 0);
     }

   call ("header_bob");
   do
     {
	thread_was_collapsed = is_thread_collapsed ();
	uncollapse_thread ();
	t_size = thread_size ();
	for (i = 1; i <= t_size; i++)
	  {
	     set_header_flags (get_header_flags () xor HEADER_READ);
	     if (i < t_size) { ifnot (header_down (1)) throw (CursorMovementError); }
	  }
	if (thread_was_collapsed) collapse_thread ();
     }
   while (header_down (1));
}
%}}}
definekey ("Enhanced->invert_read_unread",    "\ei",  "article");
